﻿using System;
using System.IO;

using Microsoft.Xna.Framework.Graphics;

using Syntriax.Engine.Core;
using Syntriax.Engine.Core.Debug;
using Syntriax.Engine.Core.Serialization;
using Syntriax.Engine.Integration.MonoGame;
using Syntriax.Engine.Serializers.Yaml;

Universe universe = new();

ISerializer serializer = new YamlSerializer();

ILogger logger = new FileLogger($"Logs/{DateTime.UtcNow:yyyy-MM-dd_HH-mm-ss-ffffff}.log");

#if DEBUG
logger = new LoggerWrapper(logger, new ConsoleLogger());
#endif

universe.InstantiateUniverseObject().SetUniverseObject("Logger").BehaviourController.AddBehaviour<LoggerContainer>().Logger = ILogger.Shared = logger;

string settingsPath = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "settings.yaml");
ClientConfiguration clientConfiguration = GetOrCreateConfiguration(serializer, logger, settingsPath);

if (clientConfiguration.Host)
    Pong.PongUniverse.ApplyPongServer(universe, clientConfiguration.HostPort);

Pong.PongUniverse.ApplyPongClient(universe, clientConfiguration.ServerAddress, clientConfiguration.ServerPort);
Pong.PongUniverse.ApplyPongUniverse(universe);

universe.InstantiateUniverseObject().SetUniverseObject("Desktop HO")
    .BehaviourController.AddBehaviour<KeyboardInputs>();

using MonoGameWindow monoGameWindow = new(universe);

monoGameWindow.Graphics.PreferredBackBufferWidth = clientConfiguration.windowWidth;
monoGameWindow.Graphics.PreferredBackBufferHeight = clientConfiguration.windowHeight;
monoGameWindow.Graphics.GraphicsProfile = GraphicsProfile.HiDef;

universe.FindBehaviour<Syntriax.Engine.Network.INetworkCommunicatorClient>()?
    .OnConnectionEstablished.AddOneTimeListener(
        (sender, connection) => monoGameWindow.Window.Title = $"Client {connection.Id}"
    );

monoGameWindow.Run();

static ClientConfiguration GetOrCreateConfiguration(ISerializer serializer, ILogger logger, string settingsPath)
{
    ClientConfiguration clientConfiguration;

    try
    {
        clientConfiguration = serializer.Deserialize<ClientConfiguration>(File.ReadAllText(settingsPath));
        logger.Log(clientConfiguration, $"Configuration is successfully read");
    }
    catch (FileNotFoundException)
    {
        clientConfiguration = new();
        logger.Log(clientConfiguration, $"Configuration does not exist");
        File.WriteAllText(settingsPath, serializer.Serialize(clientConfiguration));
    }
    catch (Exception exception)
    {
        clientConfiguration = new();
        logger.LogError(clientConfiguration, $"Error loading configuration");
        logger.LogException(clientConfiguration, exception);
        File.WriteAllText(settingsPath, serializer.Serialize(clientConfiguration));
    }

    logger.Log(clientConfiguration, $"Creating new configuration file at {settingsPath}");
    return clientConfiguration;
}
